/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 Sept  2004
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/


#ifndef INTRANSIT_H
#define INTRANSIT_H


/*
The InTransit object stores all packets that are currently being transmitted in the medium 
(i.e in the process of transmission , in transit or in the process of transmission).
 Each packet object stored has the time that transmission began, the length of the packet 
and the point in space that the transmission began from. Given a specific topology 
object, the transmission coding parameters and the InTransit object, the position of any packet
at any time may be determined.
Hence, the state of the medium at any point in time or space may be determined.

InTransit object is a singleton class
*/

class InTransit
{

//Default packet object storage mechanism is a queue.

private:
	//Packet * current;
	int debug;


public:
	Queue *q;

public:
	InTransit(int, int);
	virtual ~InTransit();

	void mark(Packet *);
	void dump();
	void add(Packet *);
	void remove(Packet *);
	void clean(Packet *);
	int numonline();
	inline void dprintf(char * format, ...);
	inline void dtprintf(char * format, ...);

	void dumpcheckpoint();
	int readcheckpoint(double *, int *);	
};


#endif

