/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 Sept  2004
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include <packet.h>
#include <eventhandler.h>
#include <queue.h>
#include <scheduler.h>
#include <state.h>
#include <rng.h>
#include <trgen.h>
#include <trdump.h>
#include <txsched.h>
#include <intransit.h>
#include <codeset.h>
#include <debug.h>
#include <receiver.h>
#include <stateest.h>
#include <state.h>
#include <statshandler.h>


extern Scheduler * sched;
extern InTransit * it;
extern TrDump * trdump;
extern TrGen * trgen;
extern char outputfile[20];
extern Codeset *codeset;
extern Receiver *rcvr;
extern StatsHandler * stats;
extern TxSched *txsched;

extern int propSpeed;


StateEstimator::StateEstimator(int n, int a):EventHandler()
{
	A =a;
	N = n;


	debug = debugon("stateest");

	int * tempstate = new int[A*N];
	if(tempstate == NULL)
	{
		printf("Error: stateest.cc: Memory allocation failed\n");
		exit(-1);
	}
	for(int count = 0; count < A*N; count++)
		tempstate[count]=0;

}

StateEstimator::~StateEstimator()
{

}


//The event handling function of the StateEstimator. It dequeues a packet and
//sends it to the next component.

void StateEstimator::handleevent(Packet *p)
{

	dtprintf("stateest.cc: StateEstimator::handleevent(): Entered function\n");
	dtprintf("stateest.cc: StateEstimator::handleevent(): Scheduling transmission: %d.\n", p->id);
	//Schedule transmission and transmit
	txsched->process(p);

}




void StateEstimator::dumpstats()
{

}


	


void StateEstimator::process(Packet *p)
{
	dtprintf("stateest.cc: StateEstimator::process(): Entered function: %d\n", p->id);
	double x = p->depx;

	double chiplength = (1.0/codeset->chiprate) * propSpeed;	
	double bitlength = codeset->N * chiplength;
	int A = codeset->A;
	int N = codeset->N;

//Determine the estimation point	
	double estpt= 0;
	if((topology == TOPOLOGY_STAR) || (topology == TOPOLOGY_BUS))
	{
		dtprintf("stateest.cc: StateEstimator::process(): Star topology\n");
		//If the transmission point is at a distance x from the coupler on the transmit fiber, then
		//it is assumed that the estimation point is on the receive fiber at a distance x from the coupler.
		estpt = -x;
	}


	dtprintf("stateest.cc: StateEstimator::process(): Estimation point: %.6f\n", estpt);
	dtprintf("stateest.cc: StateEstimator::process(): Transmission point: %.6f\n", x);
	dtprintf("stateest.cc: StateEstimator::process(): Estimation time: %.12f\n", sched->currenttime);
	dtprintf("stateest.cc: StateEstimator::process(): Estimation type: %d\n", alg);
	dtprintf("stateest.cc: StateEstimator::process(): Packet id: %d\n", p->id);
	dtprintf("stateest.cc: StateEstimator::process(): this id: %d\n", this);
	dtprintf("stateest.cc: StateEstimator::process(): Packet: %d\n", p);


//Perfect state estimation
//Perfect state estimation is defined as state estimation which returns the
//state
//Perfect state estimation can be done at any point ie. on networks with normalized propagation delay
//a =0 as well as a>0
	if(alg == STATEEST_PERFECT)
	{


//Note: Perfest state estimation can be done at any point i.e. prop delay a =0 or a>0
		int * truestate = calcstate(estpt);


//The ranging operation.
//The ranging operation introduces a delay such that the state estimate can be used for the scheduling.
		int chipoffset = calcchipoffset(estpt, x);			
		int * tstate = codeset->LeftShiftCodeword(truestate, chipoffset);	
		delete truestate;			
		

		dtprintf("stateest.cc: StateEstimator::estimatestate: True state: ");
		for (int count1 =0; count1 < A; count1++)
		{
			for (int count =N - 1; count >=0 ; count--)
			{
				dprintf("%d ", tstate[count1*N + count]);
			}
		}		
	

		p->state = tstate;
		p->ep = estpt + chipoffset *chiplength; 
	
		//Transmission schedule
		sched->scheduleevent(this, p, 0);	
		//handleevent(p);
		return;
	}



//Must be realistic state estimation
//Either continuous or on-demand


	int sample =0;
	int * totalstate = new int[A*N];
	if(totalstate == NULL)
	{
		printf("Error: stateest.cc: Memory allocation failed\n");
		exit(-1);
	}
	for (int count = 0; count < A*N; count++)
		totalstate[count] = 0;

//Sample means algorithm begins
	for(int count = 0; count < ne; count++)
	{
		dtprintf("stateest.cc: StateEstimator::process(): State sample number: %d\n", count);

//This code emulates the two types of state estimation by changing the point where the state is estimated.
//If state t_e= -10, i.e 10 secs before the packet arrived (t=0), then that state should be avaialble on the line at a distance equal to the distance beyond the estimation travelled in 10 secs.
//This effectively converts the t_e paramter of the state estimation algorithm to the correct point on the receive fiber to
//For continuous state estimation, the state samples are collected from points beyond the estimation point.
//For on demand state estimation, the state samples are taken from points before the estiamtion point. 

//For continuous
//Sampling parameters: Samples are collected continuously and dumped to a buffer whose size is at least n_e
//t_s = 0 
//n_s = inf
//b_s = n_e;
//Estimation paramters 
//n_s t_b < t_e < t_a (t_e is set to be n_s t_b secs before packet arrival)
//n_e = (Class variable)
//del_e = t_b (It is assumed that the estimation algorithm can execute in a very short time (less than a bit time). Therfore
//the latest state estimate is available.).


//For on-demand
//Sampling parameters: Samples are collected on packet arrival and dumped to a buffer whose size is at least n_e
//t_s = t_a 
//n_s = n_e
//b_s = n_e;
//Estimation paramters 
//t_e = t_a (Estimation starts on packet arrival).
//n_e = (Class variable)
//del_e = del_a (Estimation occurs when packet arrives).



//Continuous: Choose the state readings that are further off on the line i.e. before the transmission time.
//read it from.
		if((alg == STATEEST_CONTINUOUS || alg == STATEEST_REDUCED))
		{
			dtprintf("stateest.cc: StateEstimator::process(): Continuous state estimation.\n");
			estpt = (estpt - count *bitlength);
		}
//On demand: Choose the state readings that are closer on the line, i.e. starting at this time.
		else if(alg == STATEEST_ONDEMAND)
		{
			dtprintf("stateest.cc: StateEstimator::process(): On demand state estimation.\n");
			estpt = (estpt + count *bitlength);
		}


		
		dtprintf("stateest.cc: StateEstimator::process(): Before calcinstobservation 1.\n");
//The sampling operation. State samples are collected.
		int * lstate;
		//lstate = calcstateobservation(estpt - count *bitlength);
		lstate = calcstateobservation(estpt);

		if(alg == STATEEST_REDUCED)
		{
			hardlimit(lstate, 2); 
		}


//The ranging operation.
//The ranging operation introduces a delay such that the state estimate can be used for the scheduling.
		dtprintf("stateest.cc: StateEstimator::process(): Before calstateobservation.\n");
		int chipoffset = calcchipoffset(estpt, x);			
		dprintf("stateest.cc: StateEstimator::process(): Chip offset: %d\n", chipoffset);
		int * state = codeset->LeftShiftCodeword(lstate, chipoffset);	
		delete lstate;			


		dtprintf("stateest.cc: StateEstimator::estimatestate: State: ");
		for (int count1 =0; count1 < A; count1++)
		{
			for (int count2 =N - 1; count2 >=0 ; count2--)
			{
				dprintf("%d ", state[count1*N + count2]);
			}
		}
		dprintf("\n");



//The sample means algorithm.
		if(sampling == NO)
		{

		dtprintf("stateest.cc: StateEstimator::process(): Before totalling 1.\n");
/*		Packet *np = it->q->head;
		while(np != NULL) 
		{
			np->checkdata();
			np=np->next;
		}
*/
	

			for (int count1 =0; count1 < A; count1++)
			{
				for (int count2 =0; count2 < N; count2++)
				{
			 		int t = count1*N +count2;
					totalstate[count1*N + count2] += state[count1*N + count2];
					dtprintf("stateest.cc: StateEstimator::process(): t : %d.\n", t);
				}
			}
		dtprintf("stateest.cc: StateEstimator::process(): After totalling 1.\n");
/*		np = it->q->head;
		while(np != NULL) 
		{
			np->checkdata();
			np=np->next;
		}
*/		dtprintf("stateest.cc: StateEstimator::process(): After totalling 2.\n");


		}
		else
		{
			//One sample per intantaneous state reading
	 		totalstate[sample] += state[sample];
			int sample2 = (sample + (N/3)) % N;
	 		totalstate[sample2] += state[sample2];
			int sample3 = (sample + (2*N/3)) % N;
	 		totalstate[sample3] += state[sample3];
			sample++;
			if(sample >= A *N)
				sample = 0;
		}		
		delete state;


	}






//The sample means algorithm.
	p->state = new int[A*N];
	if(p->state == NULL)
	{
		printf("Error: stateest.cc: Memory allocation failed\n");
		exit(-1);
	}
	//Divide state by window/2
	if(sampling == NO)
	{
		


		for (int count1 =0; count1 < A; count1++)
		{
			for (int count2 =0; count2 < N; count2++)
			{
		 		double chip = ((double)totalstate[count1*N + count2])/(ne/2.0);
				//If chip is a fraction, then round off to the nearest integer
				int c = (int)(chip + 0.5);	
				p->state[count1*N + count2] = c;

			}
		}
	}
	else
	{

		for (int count1 =0; count1 < A; count1++)
		{
			for (int count2 =0; count2 < N; count2++)
			{
		 		double chip = ((double)totalstate[count1*N + count2]);
				//If chip is a fraction, then round off to the nearest integer
				int c = (int)(chip + 0.5);	
				p->state[count1*N + count2] = c;

			}
		}



	}

	//TBD: Need to add the offset due to ranging (chipoffset *chiplength);
	p->ep = estpt;



	dtprintf("stateest.cc: StateEstimator()::process(): Estimated state: ");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count2 =N - 1; count2 >=0 ; count2--)
		{
			dprintf("%d ", p->state[count1*N + count2]);
		}
	}
	dprintf("\n");




//Add the delay 
	double chiptime = (1.0/codeset->chiprate);	
	double rdelay;
	if(alg == STATEEST_ONDEMAND)
	{

		//If state estimation is on demand, add the state estimation delay;
		rdelay = ne * codeset->N *chiptime;
	}
	else if ((alg == STATEEST_CONTINUOUS) || (alg == STATEEST_REDUCED))
	{
		rdelay = 0;
	}



//Note that the delay occurs before the transmission schedule algorithm 
//is called.

	sched->scheduleevent(this, p, rdelay);	

	delete totalstate;
	//delete tstate;
}



int * StateEstimator::calcstate(double ex)
{
//State is estimated at point ex
//tx is the transmission point
//The node needs to know the phase (chip offset) difference between the two  so that it can schedule properly.

//tx is used to calculate the phase difference from ex

	it->dump();
	dtprintf("stateest.cc: StateEstimator::estimatestate(): A, N: %d %d %d\n", A, N, A*N);


	double timeoffset = sched->currenttime;
	double chiptime = 1.0/codeset->chiprate;
	int chipoffset = int((timeoffset/chiptime) + 0.5);

	chipoffset = chipoffset % (int) N;	
	dtprintf("stateest.cc: StateEstimator::estimatestate(): Estimation time offset from normalized time (chips): %d\n", chipoffset);

	dtprintf("stateest.cc: StateEstimator::estimatestate(): Estimation point : %.6f\n", ex);
	//printf("\n");



	int * estate = new int[A*N];
	if(estate == NULL)
	{
		printf("Error: stateest.cc: Memory allocation failed\n");
		exit(-1);
	}
	dtprintf("stateest.cc: StateEstimator::estimatestate(): estate: %d\n", estate);
	it->dump();

	Packet *p = it->q->head;
	for(int count = 0; count < A*N; count++)
		estate[count]=0;
	it->dump();
	while(p != NULL) 
	{
		dtprintf("stateest.cc: StateEstimator::estimateetate: Checking packet: %d\n", p->id);
	
		double chiplength = (1.0/codeset->chiprate) * propSpeed;	
		double packethead = 0.0;
		double packettail = 0.0;
		double x1 = ex;
		double x2 = ex + N*chiplength;
		p->calcboundary(&packethead, &packettail);
		
//Packet does not overlap
		if(packettail <=x1) 
		{
			p=p->next;
			continue;
		}
//All packets after this cannot overlap, hence break out
		if(packethead >= x2)
		{
			p=p->next;
			continue;
		}

		int allones =1;
		int * state = p->calcstate(ex, allones);
		dtprintf("stateest.cc: StateEstimator::estimateetate: Packet in window: %d\n", p->id);
		dtprintf("stateest.cc: StateEstimator::estimatestate: Code: ");
		for (int count1 =0; count1 < A; count1++)
		{
			for (int count =N - 1; count >=0 ; count--)
			{
				dprintf("%d ", state[count1*N + count]);
			}
		}
		
		dprintf("\n");
		for (int count = 0; count < A*N; count++)
			estate[count]+=state[count];		

		delete state;
		p = p->next;
	}

	return estate;
}


int  * StateEstimator::calcstateobservation(double ex)
{

	dtprintf("stateest.cc: StateEstimator::estimateetate: Calc inst state\n");



	int * lstate = new int [A*N];
	if(lstate == NULL)
	{
		printf("Error: stateest.cc: Memory allocation failed\n");
		exit(-1);
	}
	for (int count = 0; count < A*N; count++)
		lstate[count]=0;
	Packet *p = it->q->head;;
	while(p != NULL )
	{
		dtprintf("stateest.cc: StateEstimator::estimateetate: Checking packet: %d\n", p->id);
	
		double chiplength = (1.0/codeset->chiprate) * propSpeed;	
		double packethead = 0.0;
		double packettail = 0.0;
		double x1 = ex;
		double x2 = ex + N*chiplength;
		p->calcboundary(&packethead, &packettail);
		
//Packet does not overlap
		if(packettail <=x1) 
		{
			p=p->next;
			continue;
		}

//All packets after this cannot overlap, hence break out
		if(packethead >= x2)
		{
			p=p->next;
			continue;
		}
		int allones = 0;
		int * state = p->calcstate(ex, allones);

		dtprintf("stateest.cc: StateEstimator::estimateetate: Packet in window: %d\n", p->id);
		dtprintf("stateest.cc: StateEstimator::estimatestate: Code: ");
		for (int count1 =0; count1 < A; count1++)
		{
			for (int count =N - 1; count >=0 ; count--)
			{
				dprintf("%d ", state[count1*N + count]);
			}
		}
		
		dprintf("\n");
		for (int count = 0; count < A*N; count++)
			lstate[count]+=state[count];		
		


		delete state;
		p=p->next;
	}


	return lstate;
}

int StateEstimator::calcchipoffset(double ex, double tx)
{

//The state at the estimation point will not be bit synchronized with the state at 
//the transmission point
//Hence the phase difference needs to be determined.
	
	dtprintf("stateest.cc: StateEstimator::estimatestate(): Finding offset\n");
	double chiplength = (1.0/codeset->chiprate) * propSpeed;	
	double offset = tx - ex;

	int chipoffset =0;
	if (offset >= 0)
		chipoffset = int((offset/chiplength) + 0.5);
	else
		chipoffset = int((offset/chiplength) - 0.5);

	chipoffset = chipoffset % (int) N;	
	if(chipoffset < 0)
		chipoffset = -chipoffset;
		
	dtprintf("stateest.cc: StateEstimator::estimatestate(): Estimation distance offset (chips): %d\n", chipoffset);
	dtprintf("stateest.cc: StateEstimator::estimatestate(): Estimation point : %.6f\n", ex);
	//printf("\n");
	return chipoffset;
}




int StateEstimator::dumpstate(int *s1)
{

	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =N - 1; count >=0 ; count--)
		{
			dprintf("%d ", s1[count1*N + count]);
		}
	}

	return 0;	

}



void StateEstimator::hardlimit(int *s1, int limit)
{

	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =N - 1; count >=0 ; count--)
		{
			if(s1[count1*N + count] > limit)
				s1[count1*N + count] = limit;
		}
	}

}


