
/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 Sept  2004
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/


#ifndef TRGEN_H
#define TRGEN_H


/*
The traffic generator class. This object reads the input traffic file 
and generates the packet objects and passes them on to the required 
processing objects at the right time.
*/


class TrGen :public EventHandler
{
private:

public:

	int intarrdist;
	int pktsizedist;
	int topologydist;

	//For bimodal packet size distribution
	double fracshortpkt;

	double lastarrival;
	double avginterarrival;
	double avgpktsize;
	double varpktsize;
	double offeredload;

	double avgdistance;
	double avgdc;
	double numnodes;

	double *dist;

	//Used to collect information for calculating offerred load when using trace files
	double lasttime;
	double initialtime;
	double totalbits;
	double totaltime;

	int totalpkts;
	int done;
	int modulation;

	int steadystatepktcnt;
	int steadystatebytecnt;

	int pktcnt;
	int bytecnt;
	int id;

	RNG *psrng;
	RNG *iarng;
	RNG *distrng;
	
	FILE *fin;
	FILE *fout;
	TrGen(int, int, double, double, int);
	virtual ~TrGen();
	virtual void handleevent(Packet *);		
	virtual void read();		
	void dumpstats();
	double genpacketsize();
	double genarrivaltime();
	int gendistance();

};


enum {TRAFFIC_EXP=0, TRAFFIC_UNIFORM, TRAFFIC_DET, TRAFFIC_TM, TRAFFIC_FILE, TRAFFIC_NORMAL, TRAFFIC_PARETO, TRAFFIC_BM};
enum {TOPOLOGY_EXP=0, TOPOLOGY_UNIFORM, TOPOLOGY_DET};

#endif

