/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 Sept  2004
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include <packet.h>
#include <eventhandler.h>
#include <queue.h>
#include <scheduler.h>
#include <state.h>
#include <rng.h>
#include <trgen.h>
#include <trdump.h>
#include <txsched.h>
#include <intransit.h>
#include <codeset.h>
#include <debug.h>
#include <receiver.h>
#include <transmitter.h>
#include <stateest.h>
#include <state.h>
#include <statshandler.h>


extern Scheduler * sched;
extern InTransit * it;
extern TrDump * trdump;
extern TrGen * trgen;
extern char outputfile[20];
extern Codeset *codeset;
extern Receiver *rcvr;
extern StateEstimator *stateest;
extern StatsHandler * stats;

extern int propSpeed;
extern int steadystate;
extern Transmitter *trn;

TxSched::TxSched():EventHandler()
{
	debug = debugon("txsched");
	dropped = 0;

}

TxSched::~TxSched()
{
}


void TxSched::process(Packet *p)
{


}

void TxSched::defer(Packet *p)
{
//TBD: Add code for defer modes:
//	- Non persistent
//	- 1-persistent
//	- p-persistent


//TBD: Add code for Interference detection

//TBD: Add code for a backoff algorithm (Binary exponential backoff, BLAM)
	dtprintf("txsched.cc: TxSched::defer(): Defering: %d\n", p->defers);

	if(deferalg == DEFER_NODEFER)
	{
		//Default defer mode
		//Delete the packet and log it as dropped
		p->status = PACKET_DROPPED;
		p->departuretime =0;
		if(steadystate == 1)
			dropped++;
		trdump->write(p);
		delete p;
	}
	else if(deferalg == DEFER_BEB)
	{

		double delay;
		p->defers++;
		int backofflimit = 16;
		if(p->defers > backofflimit)
		{
			p->departuretime =0;
			p->status = PACKET_DROPPED;
			if(steadystate == 1)
				dropped++;
			trdump->write(p);
			delete p;
			return;
		}
		double slottime = trgen->avgpktsize *8 * (codeset->N)/(codeset->chiprate);

		int minwindow = 1;
		int backofffactor =2;
		
		int currentwindow = minwindow * (int) pow(backofffactor, p->defers);
		int numslots  = (int) ((random()/(pow(2, 31) - 1))*currentwindow);
		delay = slottime * numslots;		
		dtprintf("txsched.cc: TxSched::defer(): Packet id:  %d\n", p->id);
		dtprintf("txsched.cc: TxSched::defer(): Current window: %d\n", currentwindow);
		dtprintf("txsched.cc: TxSched::defer(): Number of delay slots: %d\n", numslots);
		dtprintf("txsched.cc: TxSched::defer(): Number of defers: %d\n", p->defers);
		p->status= PACKET_DEFERED;
		sched->scheduleevent(this, p, delay);
	}
}

//The event handling function of the TxSched. It dequeues a packet and
//sends it to the next component.
void TxSched::handleevent(Packet *p)
{


	dtprintf("txsched.cc: TxSched::handleevent(): Entered function\n");
}



void TxSched::dumpstats()
{
	stats->droppedpkts = dropped;
}



TxSched_Aloha::TxSched_Aloha():TxSched()
{
}

TxSched_Aloha::~TxSched_Aloha()
{
}


void TxSched_Aloha::process(Packet *p)
{

	//Schedule the event for cwtransmission immediately
	dtprintf("trdump.cc: TxSched_Aloha::process(): Scheduling packet: %d\n", p->id); 		
	sched->scheduleevent(this, p, 0);	

}


//The event handling function of the TxSched_Aloha. It dequeues a packet and
//sends it to the next component.
void TxSched_Aloha::handleevent(Packet *p)
{

	dtprintf("txsched.cc: TxSched_Aloha::handleevent(): Entered function\n");
	dprintf("\n");
	p->departuretime = sched->currenttime; 	 
	p->offset =0;	
	p->status = PACKET_INTRANSIT;
	it->add(p);


	trdump->process(p);
}



void TxSched_Aloha::dumpstats()
{

	stats->droppedpkts = dropped;
}


TxSched_Selfish::TxSched_Selfish(int n, int a, int W, int k):TxSched()
{

	N = n;
	A = a;
	w = W;
	K= k;
}

TxSched_Selfish::~TxSched_Selfish()
{
}


void TxSched_Selfish::process(Packet *p)
{
	dtprintf("txsched.cc: TxSched_Selfish::process(): Scheduling packet: %d\n", p->id);
	dtprintf("txsched.cc: TxSched_Selfish::process(): Codeword :\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =N-1; count >=0; count--)
		{
			dprintf("%d ", codeset->set[p->codeword *(A*N) + count1*N + count]);
		}
		dprintf("\n");
	}
	dprintf("\n");

	int * state;
	state = p->state;
	dtprintf("txsched.cc: TxSched_Selfish::process(): Estimated state:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =N -1; count >=0; count--)
		{
			dprintf("%d ", state[count1*N + count]);
		}
		dprintf("\n");
	}


	int offsetcount =0;
	int offset[w * N];


	for (int count3 =0; count3 < N; count3++)
	{
		int tdelay = count3;	
		int * shifted = codeset->RightShiftCodeword(p->codeword, tdelay);
		
		int selfish = 0;		
		for (int count4 =0; count4 < A; count4++)	
		{	
			for (int count5 =N-1; count5 >= 0; count5--)	
			{

				//Check for selfishness
				if((state[count4*N + count5] == 0) && (shifted[count4 *N + count5] == 1))
				{
					selfish = 1;
					offset[offsetcount] = tdelay;
					offsetcount++;
					break;
				}
			}
			if(selfish==1)
				break;
		}

		delete shifted;
	}








	if (offsetcount != 0)
	{

//Choose one of the available offsets randomly.

		int rnd = (int)(floor((random()/(pow(2, 31) -1)) *offsetcount));
		int off = offset[rnd];

		double timeoffset = sched->currenttime;
		double chiptime = 1.0/codeset->chiprate;
		int chipoffset = int((timeoffset/chiptime) + 0.5);
		chipoffset = chipoffset % (int) N;	
		int normoffset  = (chipoffset + off) % N;
	
		dtprintf("txsched.cc: TxSched_Selfish::handle_event:: Sending at offset (relative to estimation time): %d\n", off);
		dtprintf("txsched.cc: TxSched_Selfish::handle_event:: Sending at offset (relative to normalized time): %d\n", normoffset);
		


		double delay = (double)off * chiptime; 	
		p->departuretime = sched->currenttime + delay;


		p->offset = off;
		p->status = PACKET_INTRANSIT;
		//it->add(p);
		dtprintf("txsched.cc: TxSched_Selfish::handle_event::Departure time: %.12f\n", p->departuretime );
		sched->scheduleevent(this, p, delay);	
	}
	else
	{
		dtprintf("txsched.cc: TxSched::handle_event::Cannot schedule packet: %d\n", p->id);
		defer(p);
	}



}


//The event handling function of the TxSched_Selfish. It dequeues a packet and
//sends it to the next component.
void TxSched_Selfish::handleevent(Packet *p)
{

		dtprintf("txsched.cc: TxSched_Selfish::handleevent(): Entered function\n");
		dprintf("\n");
		it->add(p);
		trdump->process(p);
}



void TxSched_Selfish::dumpstats()
{
	stats->droppedpkts = dropped;

}


TxSched_Threshold::TxSched_Threshold(int n, int a, int W, int k, double thparam):TxSched()
{

	N = n;
	A =a;
	w = W;
	K=k;
	th = thparam* N *A;
}

TxSched_Threshold::~TxSched_Threshold()
{
}


void TxSched_Threshold::process(Packet *p)
{
//Overlap section scheduling: Schedule if selfish and number of overlaps is less than the threshold

	//Schedule the event for whenever it passes the headend
	dtprintf("trdump.cc: TxSched_Threshold::process(): Scheduling a packet: %d\n", p->id); 		
	dtprintf("trdump.cc: TxSched_Threshold::process(): Threshold: %.6f\n", th); 		
	
	dtprintf("txsched.cc: TxSched_Threshold::process(): Code: %d\n", p->codeword);
	dtprintf("txsched.cc: TxSched_Threshold::process(): Code:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =N-1; count >=0; count--)
		{
			dprintf("%d ", codeset->set[p->codeword *(A*N) + count1*N + count]);
		}
		dprintf("\n");
	}
	dprintf("\n");

	/*
	//Sanity checking code for perfect state estimation
	int overlapsanity=0;
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =N - 1; count >=0 ; count--)
		{
			if(p->state[count1*N + count] > 1)
				overlapsanity++;
		}
	}
	if(overlapsanity >= th)
	{
		printf("Error: Threshold exceeded\n");
		exit(-1);
	}
	//End of sanity check
	*/
	int chips = 0;
	int n0=0, n1=0, nov=0;
	dtprintf("txsched.cc: TxSched_Threshold::process(): State:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =N-1; count >=0; count--)
		{
			dprintf("%d ", p->state[count1*N + count]);
			chips+=p->state[count1*N + count];
			if(p->state[count1*N+count] == 0)
				n0++;
			else if(p->state[count1*N+count] == 1)
				n1++;
			else
				nov++;
		}
		dprintf("\n");
	}
	int numonline =  chips/w;
	dtprintf("txsched.cc: TxSched_Threshold::process(): Number on line: %d: %.6f\n", numonline, p->depx ); 		
	dtprintf("txsched.cc: TxSched_Threshold::process(): Zeros: %d Ones: %d Overlaps: %d\n", n0, n1, nov);

	int *state = p->state;
	double threal = th;


	int offsetcount =0;
	int offset[w * N];
	for (int count3 =0; count3 < N; count3++)
	{
		int overlaps = 0;
		int tdelay = count3;	
		int * shifted = codeset->RightShiftCodeword(p->codeword, tdelay);
		
		int selfish = 0;		
		for (int count4 =0; count4 < A; count4++)	
		{	
			for (int count5 =N-1; count5 >= 0; count5--)	
			{

				//Check for selfishness
				if((state[count4*N + count5] == 0) && (shifted[count4 *N + count5] == 1))
				{
					selfish = 1;
				}
				if((state[count4*N + count5]  +  shifted[count4*N + count5]) > 1) 
					overlaps++;
			}
		}

		//dtprintf("Overlaps: %d Threshold: %.6f\n", overlaps, threal);


		if((overlaps < (int)threal)   && (selfish == 1))
		{
			offset[offsetcount] = tdelay;
			offsetcount++;
		}
		delete shifted;
	}







	if (offsetcount != 0)
	{

//Choose one of the available offsets randomly.

		int rnd = (int)(floor((random()/(pow(2, 31) -1)) *offsetcount));
		int off = offset[rnd];
		//int off = offset[offsetcount -1];

		dtprintf("txsched.cc: TxSched::handle_event:: %d Choosing offset: %d\n", p->codeword, off );
		dtprintf("txsched.cc: TxSched::handle_event:: id %d\n", p->id);
		double chiptime = (1.0/p->chiprate);
		double delay = (double)off * chiptime; 	
		p->departuretime = sched->currenttime + delay;
		p->offset = off;
		p->status = PACKET_INTRANSIT;
		//it->add(p);
		dtprintf("txsched.cc: TxSched::handle_event::Departure time: %.12f\n", p->departuretime );
		dtprintf("txsched.cc: TxSched::handle_event::Packet departing %d at %.6f\n", p->id, p->depx);
		
		sched->scheduleevent(this, p, delay);	
	}
	else
	{
		dtprintf("txsched.cc: TxSched::handle_event::Cannot schedule packet: %d\n", p->id);
		defer(p);
	}

	dprintf("\n");


}


//The event handling function of the TxSched_Threshold. It dequeues a packet and
//sends it to the next component.
void TxSched_Threshold::handleevent(Packet *p)
{
	dtprintf("txsched.cc: TxSched_Threshold::handleevent(): Entered function\n");

//TBD: Implement defers for other txscheduling algs

	if(p->status == PACKET_INTRANSIT)
	{
		it->add(p);
		trdump->process(p);
	}
	else if(p->status == PACKET_DEFERED)
	{
		//Defered packet, send to transmiiter
		trn->process(p);
	}
}


void TxSched_Threshold::dumpstats()
{
	stats->droppedpkts = dropped;

}

TxSched_OverlapSection::TxSched_OverlapSection(int n, int a, int W, int k, double thparam):TxSched()
{

	N = n;
	A =a;
	w = W;
	K=k;
	th = thparam* N;
}

TxSched_OverlapSection::~TxSched_OverlapSection()
{
}


void TxSched_OverlapSection::process(Packet *p)
{
//Overlap section scheduling: Schedule if selfish and number of overlaps is less than number of ones.

	//Schedule the event for whenever it passes the headend
	dtprintf("trdump.cc: TxSched_OverlapSection::process(): Scheduling a packet\n"); 		
	dtprintf("trdump.cc: TxSched_OverlapSection::process(): OverlapSection: %.6f\n", th); 		

	dtprintf("txsched.cc: TxSched_OverlapSection::process(): Code:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =N-1; count >=0; count--)
		{
			dprintf("%d ", codeset->set[p->codeword *(A*N) + count1*N + count]);
		}
		dprintf("\n");
	}


	int * state = p->state;
	dtprintf("txsched.cc: TxSched_OverlapSection::process(): Estimated state:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =N-1; count >=0; count--)
		{
			dprintf("%d ", state[count1*N + count]);
		}
		dprintf("\n");
	}

	int offsetcount =0;
	int offset[w * N];

	for (int count3 =0; count3 < N; count3++)
	{
		int overlaps = 0;
		int ones = 0;
		int tdelay = count3;	
		int * shifted = codeset->RightShiftCodeword(p->codeword, tdelay);
		
		int selfish = 0;		
		for (int count4 =0; count4 < A; count4++)	
		{	
			for (int count5 =N-1; count5 >= 0; count5--)	
			{

				//Check for selfishness
				if((state[count4*N + count5] == 0) && (shifted[count4 *N + count5] == 1))
					selfish = 1;
				if((state[count4*N + count5] + shifted[count4*N + count5]) >1) 
					overlaps++;
				if ((state[N*count4 + count5] + shifted[N*count4+count5]) == 1) 
					ones++; 		
			}
		}

		dtprintf("Overlaps: %d ones: %d\n", overlaps, ones);

		if((overlaps < ones) && (selfish == 1))
		{
			offset[offsetcount] = tdelay;
			offsetcount++;
		}

		delete shifted;
	}







	if (offsetcount != 0)
	{

//Choose one of the available offsets randomly.

		int rnd = (int)(floor((random()/(pow(2, 31) -1)) *offsetcount));
		int off = offset[rnd];

		dtprintf("txsched.cc: TxSched::handle_event:: %d Choosing offset: %d\n", p->codeword, off );
		dtprintf("txsched.cc: TxSched::handle_event:: id %d\n", p->id);
		double chiptime = (1.0/p->chiprate);
		double delay = (double)off * chiptime; 	
		p->departuretime = sched->currenttime + delay;
		p->offset = off;
		p->status = PACKET_INTRANSIT;
		//it->add(p);
		dtprintf("txsched.cc: TxSched::handle_event::Departure time: %.12f\n", p->departuretime );

		sched->scheduleevent(this, p, delay);	
	}
	else
	{
		dtprintf("txsched.cc: TxSched::handle_event::Cannot schedule packet: %d\n", p->codeword);
		defer(p);
	}

	dprintf("\n");
	dprintf("\n");


}


//The event handling function of the TxSched_OverlapSection. It dequeues a packet and
//sends it to the next component.
void TxSched_OverlapSection::handleevent(Packet *p)
{
	it->add(p);
	trdump->process(p);
	dtprintf("txsched.cc: TxSched_OverlapSection::handleevent(): Entered function\n");
}



void TxSched_OverlapSection::dumpstats()
{
	stats->droppedpkts = dropped;

}

TxSched_OverlapSize::TxSched_OverlapSize(int n, int a, int W, int k, double thparam):TxSched()
{

	N = n;
	A =a;
	w = W;
	K=k;
	th = thparam;
}

TxSched_OverlapSize::~TxSched_OverlapSize()
{
}


void TxSched_OverlapSize::process(Packet *p)
{

	//Schedule the event for whenever it passes the headend
	dtprintf("trdump.cc: TxSched_OverlapSize::process(): Scheduling a packet\n"); 		
	dtprintf("trdump.cc: TxSched_OverlapSize::process(): OverlapSize: %.6f\n", th); 		
	dtprintf("txsched.cc: TxSched_OverlapSize::process(): Code:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =N-1; count >=0; count--)
		{
			dprintf("%d ", codeset->set[p->codeword *(A*N) + count1*N + count]);
		}
		dprintf("\n");
	}


	int * state = p->state;
	dtprintf("txsched.cc: TxSched_OverlapSize::process(): Estimated state:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =N-1; count >=0; count--)
		{
			dprintf("%d ", state[count1*N + count]);
		}
		dprintf("\n");
	}



	int offset = -1;	

	if (offset != -1)
	{
		dtprintf("txsched.cc: TxSched::handle_event::Choosing offset: %d\n", offset );
		double chiptime = (1.0/p->chiprate);
		double delay = (double)offset * chiptime; 	
		p->departuretime = sched->currenttime + delay;
		p->offset = offset;
		p->status = PACKET_INTRANSIT;
		//it->add(p);
		dtprintf("txsched.cc: TxSched::handle_event::Departure time: %.12f\n", p->departuretime );
		sched->scheduleevent(this, p, delay);	
	}
	else
	{
		dtprintf("txsched.cc: TxSched::handle_event::Cannot schedule packet: %d\n");
		defer(p);
	}



	dprintf("\n");

}


//The event handling function of the TxSched_OverlapSize. It dequeues a packet and
//sends it to the next component.
void TxSched_OverlapSize::handleevent(Packet *p)
{
	it->add(p);
	trdump->process(p);
	dtprintf("txsched.cc: TxSched_OverlapSize::handleevent(): Entered function\n");
}


//TBD: Optimal state: Schedule such that distance to the optimal state is minimized
//This may be the optimal scheduling algorithm
//Needs to be proved.
void TxSched_OverlapSize::dumpstats()
{
	stats->droppedpkts = dropped;

}

TxSched_OptimalState::TxSched_OptimalState(int n, int a, int W, int k):TxSched()
{

	N = n;
	A = a;
	w = W;
	K= k;
	printf("Optimal transmission scheduling not implemeneted\n");
	exit(-1);
}

TxSched_OptimalState::~TxSched_OptimalState()
{
}


void TxSched_OptimalState::process(Packet *p)
{
	dtprintf("txsched.cc: TxSched_OptimalState::process(): Scheduling packet: %d\n", p->id);
	dtprintf("txsched.cc: TxSched_OptimalState::process(): Code:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =N-1; count >=0; count--)
		{
			dprintf("%d ", codeset->set[p->codeword *(A*N) + count1*N + count]);
		}
		dprintf("\n");
	}

	int * state;

	

	state = p->state;
	dtprintf("txsched.cc: TxSched_OptimalState::process(): Estimated state:\n");
	dtprintf("State:");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =N-1; count >=0; count--)
		{
			dprintf("%d ", state[count1*N + count]);
		}
		dprintf("\n");
	}


	dprintf("\n");
}


//The event handling function of the TxSched_OptimalState. It dequeues a packet and
//sends it to the next component.
void TxSched_OptimalState::handleevent(Packet *p)
{
	dtprintf("txsched.cc: TxSched_OptimalState::handleevent(): Entered function\n");
	dprintf("\n");
	trdump->process(p);
}



void TxSched_OptimalState::dumpstats()
{
	stats->droppedpkts = dropped;

}


//Admission control: Admit based on the number of users.
//Do not schedule packets		
TxSched_AdmissionControl::TxSched_AdmissionControl(int n, int a, int W, int k, double thparam):TxSched()
{

	N = n;
	A =a;
	w = W;
	K=k;
	th = thparam* N *A;
}

TxSched_AdmissionControl::~TxSched_AdmissionControl()
{
}


void TxSched_AdmissionControl::process(Packet *p)
{

	dtprintf("trdump.cc: TxSched_AdmissionControl::process(): Scheduling a packet\n"); 		
	dtprintf("trdump.cc: TxSched_AdmissionControl::process(): AdmissionControl: %.6f\n", th); 		
	
	dtprintf("txsched.cc: TxSched_AdmissionControl::process(): Code:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =N-1; count >=0; count--)
		{
			dprintf("%d ", codeset->set[p->codeword *(A*N) + count1*N + count]);
		}
		dprintf("\n");
	}


	int * state = p->state;
	dtprintf("txsched.cc: TxSched_AdmissionControl::process(): Estimated state:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =N-1; count >=0; count--)
		{
			dprintf("%d ", state[count1*N + count]);
		}
		dprintf("\n");
	}

	double threal = th;
	double chips = 0;

	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =N-1; count >=0; count--)
		{
			//No selfish check
			chips+= state[N*count1 + count];
		}
	}
	int nonline = (int)(chips/(double)codeset->w);

	if(nonline < threal)
	{
		double delay = 0;
		p->departuretime = sched->currenttime + delay;
		p->offset = 0;
		p->status = PACKET_INTRANSIT;
		//it->add(p);
		dtprintf("txsched.cc: TxSched::handle_event::Departure time: %.12f\n", p->departuretime );
		sched->scheduleevent(this, p, delay);	
	}
	else
	{
		dtprintf("txsched.cc: TxSched::handle_event::Cannot schedule packet: %d\n");
		defer(p);
	}

	dprintf("\n");



}


//The event handling function of the TxSched_AdmissionControl. It dequeues a packet and
//sends it to the next component.
void TxSched_AdmissionControl::handleevent(Packet *p)
{
	it->add(p);
	trdump->process(p);
	dtprintf("txsched.cc: TxSched_AdmissionControl::handleevent(): Entered function\n");
}



void TxSched_AdmissionControl::dumpstats()
{
	stats->droppedpkts = dropped;

}


//Distributed Perfect Scheduling
//Schedule such that the packet is not lost and no other packet on the line is lost
//Needs to know the exact codewords on the line for this.
//In real life that information may not be possible to calculate
//Hack: The information is added to the packet in the stateest process function
 
TxSched_DPS::TxSched_DPS(int n, int a, int W, int k, double thparam):TxSched()
{

	N = n;
	A =a;
	w = W;
	K=k;
	th = thparam* N *A;
}

TxSched_DPS::~TxSched_DPS()
{
}


void TxSched_DPS::process(Packet *p)
{

	//Schedule the event for whenever it passes the headend
	dtprintf("trdump.cc: TxSched_DPS::process(): Scheduling a packet\n"); 		
	dtprintf("trdump.cc: TxSched_DPS::process(): DPS: %.6f\n", th); 		
	
	dtprintf("txsched.cc: TxSched_DPS::process(): Code:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for(int count =N-1; count >=0; count--)
		{
			dprintf("%d ", codeset->set[p->codeword *(A*N) + count1*N + count]);
		}
		dprintf("\n");
	}

	dtprintf("txsched.cc: TxSched_DPS::process(): State:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =N-1; count >=0; count--)
		{
			dprintf("%d ", p->state[count1*N + count]);
		}
		dprintf("\n");
	}
	
	
	int * news  = new int [A*N];
	if(news == NULL)
	{
		printf("Error: tsched.cc: Memory allocation failed\n");
		exit(-1);
	}
	for(int count = 0; count < N; count++)
	{
	
	//Shift the transmitting codeword

		
		int * shiftedcodeword = codeset->RightShiftCodeword(p->codeword, count);	
		int overlaps = 0;


	//First check if it is destroyed
		for (int count3 =0; count3 < A; count3++)
		{
			for (int count4 =0; count4 < N; count4++)
			{
				news[N*count3 + count4] = p->state[N*count3+ count4] + shiftedcodeword[N*count3 + count4];
				if((news[N*count3 + count4] > 1 )&& ( shiftedcodeword[N*count3 + count4] == 1))
					overlaps++;
			}
		}

		if(overlaps == w)
		{
			
			dtprintf("txsched.cc: TxSched_DPS::process(): Error in transmitting  codeword\n");
			continue;	
		}

		dtprintf("txsched.cc: TxSched_DPS::process(): This packet is not destroyed\n");


	//Then check if it destorys any other codewords on the line

		//TBD: Optimize by combining loops.
		calclinecodewords(p);
		LineCodeword * linecodewords = p->linecodewords;
		overlaps = 0;
		int error = 0;
		while(linecodewords != NULL)
		{
	
			dtprintf("txsched.cc: TxSched_DPS::process(): Line codeword:\n");
			for (int count3 =0; count3 < A; count3++)
			{
				for (int count4 =N-1; count4 >=0; count4--)
				{
					dprintf("%d ", linecodewords->codeword[count3*N + count4]);
				}
				dprintf("\n");
			}
			dtprintf("txsched.cc: TxSched_DPS::process(): Counting overlaps\n");
			overlaps = 0;
			for (int count3 =0; count3 < A; count3++)
			{
				for (int count4 =0; count4 < N; count4++)
				{

					if((news[N*count3 + count4] > 1 )&& ( linecodewords->codeword[N*count3 + count4] == 1))
						overlaps++;
				}
			}
		
			dtprintf("txsched.cc: TxSched_DPS::process(): Delay: %d\n", count);
			if(overlaps == w)
			{
				
				dtprintf("txsched.cc: TxSched_DPS::process(): Error\n");
				error = 1;
				break;	
			}	
			dtprintf("txsched.cc: No error\n");

			linecodewords = linecodewords->next;
		}



		dtprintf("txsched.cc: Completed checking\n");
//If no errors then schedule
		if(error == 0)
		{
			dtprintf("txsched.cc: TxSched_DPS::process(): Newstate:\n");
			for (int count3 =0; count3 < A; count3++)
			{
				for (int count4 =N-1; count4 >=0; count4--)
				{
					dprintf("%d ", news[count3*N + count4]);
				}
				dprintf("\n");
			}
			double chiptime = (1.0/p->chiprate);
			double delay = (double)count * chiptime; 	
			p->departuretime = sched->currenttime + delay;
			p->offset = 0;
			p->status = PACKET_INTRANSIT;
			//it->add(p);
			dtprintf("txsched.cc: TxSched::handle_event::Departure time: %.12f\n", p->departuretime );
			sched->scheduleevent(this, p, delay);	
			return;
		}

	}
	dtprintf("txsched.cc: TxSched::handle_event::Cannot schedule packet: %d\n");
	defer(p);



}


//The event handling function of the TxSched_DPS. It dequeues a packet and
//sends it to the next component.
void TxSched_DPS::handleevent(Packet *p)
{
	it->add(p);
	trdump->process(p);
	dtprintf("txsched.cc: TxSched_DPS::handleevent(): Entered function\n");
}



void TxSched_DPS::dumpstats()
{
	stats->droppedpkts = dropped;

}


//Distributed perfect scheduling
//Add the codeword to the packet codeword list
//This can be used later to do perfect scheduling
void TxSched_DPS::calclinecodewords(Packet *pt)
{

//The state at the estimation point will not be bit synchronized with the state at 
//the transmission point
//Hence the phase difference needs to be determined.
	
	double tx = pt->depx;
	double ex = -tx;

	dtprintf("stateest.cc: StateEstimator::estimatestate(): Finding offset\n");
	double chiplength = (1.0/codeset->chiprate) * propSpeed;	
	double offset = tx - ex;

	int chipoffset =0;
	if (offset >= 0)
		chipoffset = int((offset/chiplength) + 0.5);
	else
		chipoffset = int((offset/chiplength) - 0.5);

	chipoffset = chipoffset % (int) N;	
	if(chipoffset < 0)
		chipoffset = -chipoffset;
		
	dtprintf("stateest.cc: StateEstimator::estimatestate(): Estimation distance offset (chips): %d\n", chipoffset);
	dtprintf("stateest.cc: StateEstimator::estimatestate(): Estimation point : %.6f\n", ex);
	//printf("\n");



	Packet *p = it->q->head;
	while(p != NULL )
	{
		dtprintf("stateest.cc: StateEstimator::estimateetate: Checking packet: %d\n", p->id);
	
		double chiplength = (1.0/codeset->chiprate) * propSpeed;	
		double packethead = 0.0;
		double packettail = 0.0;
		double x1 = ex;
		double x2 = ex + N*chiplength;
		p->calcboundary(&packethead, &packettail);
		
//Packet does not overlap
		if(packettail <=x1) 
		{
			p=p->next;
			continue;
		}

//All packets after this cannot overlap, hence break out
		if(packethead >= x2)
		{
			p=p->next;
			continue;
		}

		int allones = 1;
		int * state = p->calcstate(ex, allones);
		dtprintf("stateest.cc: StateEstimator::estimateetate: Packet in window: %d\n", p->id);
		dtprintf("stateest.cc: StateEstimator::estimatestate: Code: ");
		for (int count1 =0; count1 < A; count1++)
		{
			for (int count =N - 1; count >=0 ; count--)
			{
				dprintf("%d ", state[count1*N + count]);
			}
		}
		
		dprintf("\n");

		dtprintf("stateest.cc: StateEstimator::estimatestate: Copying codewords ");
		int * tstate = codeset->LeftShiftCodeword(state, chipoffset);
		delete state;

		LineCodeword *newlc = new LineCodeword();
		if(newlc == NULL)
		{
			printf("Error: txsched.cc: Memory allocation failed\n");
			exit(-1);
		}
		newlc->codeword = tstate;
		newlc->next = pt->linecodewords;
		pt->linecodewords = newlc;


		p=p->next;
	}
}



TxSched_MinCrossCorr::TxSched_MinCrossCorr(int n, int a, int W, int k, double thparam):TxSched()
{

	N = n;
	A =a;
	w = W;
	K=k;
	th = thparam* N *A;
}

TxSched_MinCrossCorr::~TxSched_MinCrossCorr()
{
}


void TxSched_MinCrossCorr::process(Packet *p)
{
//Overlap section scheduling: Schedule if selfish and number of overlaps is less than the threshold

	//Schedule the event for whenever it passes the headend
	dtprintf("trdump.cc: TxSched_MinCrossCorr::process(): Scheduling a packet: %d\n", p->id); 		
	dtprintf("trdump.cc: TxSched_MinCrossCorr::process(): MinCrossCorr: %.6f\n", th); 		
	
	dtprintf("txsched.cc: TxSched_MinCrossCorr::process(): Code: %d\n", p->codeword);
	dtprintf("txsched.cc: TxSched_MinCrossCorr::process(): Code:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =N-1; count >=0; count--)
		{
			dprintf("%d ", codeset->set[p->codeword *(A*N) + count1*N + count]);
		}
		dprintf("\n");
	}

	int chips = 0;
	dtprintf("txsched.cc: TxSched_MinCrossCorr::process(): State:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =N-1; count >=0; count--)
		{
			dprintf("%d ", p->state[count1*N + count]);
			chips+=p->state[count1*N + count];
		}
		dprintf("\n");
	}
	int numonline =  chips/w;
	printf("txsched.cc: TxSched_MinCrossCorr::process(): Number on line: %d: %.6f\n", numonline, p->depx ); 		

	int *state = p->state;

	//TBD: Try with hardlimited state
	//int *state = inew int [A*N] //=stateest->hardlimit(p->state, 1);
	



	int crosscorr[N];
	int selfish[N];

	for (int count3 =0; count3 < N; count3++)
	{
		int tdelay = count3;	
		int * shifted = codeset->RightShiftCodeword(p->codeword, tdelay);
	
		crosscorr[tdelay] = 0;	
		selfish[tdelay] = 0;	
		for (int count4 =0; count4 < A; count4++)	
		{	
			for (int count5 =N-1; count5 >= 0; count5--)	
			{

				//Check for selfishness
				if((state[count4*N + count5] == 0) && (shifted[count4 *N + count5] == 1))
					selfish[tdelay] = 1;
				crosscorr[tdelay]+= state[count4*N + count5]* shifted[count4*N + count5];
			}
		}

		//dtprintf("Overlaps: %d MinCrossCorr: %.6f\n", overlaps, threal);
		delete shifted;
	}

	int off = -1;
	int mincrosscorr = 9999999;
	for (int count3 =0; count3 < N; count3++)
	{
		//if((crosscorr[count3] < mincrosscorr)  && (selfish[count3] == 1))
		if((crosscorr[count3] < mincrosscorr)  && (crosscorr[count3] < w))
		{
			off = count3;
			mincrosscorr = crosscorr[count3];
		}
	}




	if(off != -1)
	{
		dtprintf("txsched.cc: TxSched::handle_event:: %d Choosing offset: %d\n", p->codeword, off );
		dtprintf("txsched.cc: TxSched::handle_event:: id %d\n", p->id);
		double chiptime = (1.0/p->chiprate);
		double delay = (double)off * chiptime; 	
		p->departuretime = sched->currenttime + delay;
		p->offset = off;
		p->status = PACKET_INTRANSIT;
		//it->add(p);
		dtprintf("txsched.cc: TxSched::handle_event::Departure time: %.12f\n", p->departuretime );
		printf("txsched.cc: TxSched::handle_event::Packet departing %d at %.6f\n", p->id, p->depx);
	
		sched->scheduleevent(this, p, delay);	
	}
	else
	{
		dtprintf("txsched.cc: TxSched::handle_event::Cannot schedule packet: %d\n", p->id);
		defer(p);
	}

	dprintf("\n");


}


//The event handling function of the TxSched_MinCrossCorr. It dequeues a packet and
//sends it to the next component.
void TxSched_MinCrossCorr::handleevent(Packet *p)
{
	it->add(p);
	dtprintf("txsched.cc: TxSched_MinCrossCorr::handleevent(): Entered function\n");
	trdump->process(p);
}
